# Copyright 1999-2014. Parallels IP Holdings GmbH. All Rights Reserved.
""" Fail2Ban configuration array management and merger. It's alive! """
import os
import re
import types
import random
import string
import operator
import functools
import ConfigParser
from glob import glob
from shlex import shlex

import plesk_log


log = plesk_log.getLogger(__name__)


class ConfigPresentation(object):
    INCLUDES_SECTION = 'INCLUDES'
    DEFINITION_SECTION = 'Definition'
    INIT_SECTION = 'Init'
    DEFAULT_SECTION = ConfigParser.DEFAULTSECT

    SUBDIR = ''

    def __init__(self, name, basedir):
        self._configs_name = name
        self._basedir = basedir
        self._presentations = {}

    def __repr__(self):
        return "<fail2ban %s of '%s' under '%s'>" % (self.__class__.__name__, self._configs_name, self._basedir)

    @classmethod
    def get_config_includes(cls, filename, skip_paths=[], deep=True):
        """ Traverses configuration files tree and returns shallow and optionally deep includes.

            In shallow mode (deep == False) returns dictionary with 'before' and 'after' keys with sets of 
            ("absolute" include path which includes basedir, include name as encountered in including config) pairs.

            In deep mode (default one) a second dictionary is returned, its values are lists of "absolute" 
            include paths for all included configs in an inclusion tree with filename as root.

            Includes are listed in [INCLUDES] section in multiline 'before' and 'after' options.

            This is how fail2ban processes includes.
        """
        parser = ConfigParser.RawConfigParser()
        parser.read(filename)

        basedir = os.path.dirname(filename)

        immediate_includes = {'before': set(), 'after': set()}
        deep_includes = {'before': [], 'after': []}

        if cls.INCLUDES_SECTION in parser.sections():
            for include_option in ('before', 'after'):
                if parser.has_option(cls.INCLUDES_SECTION, include_option):
                    includes = parser.get(cls.INCLUDES_SECTION, include_option)
                    for include in split_multiline_value(includes):
                        abs_include = include if os.path.isabs(include) else os.path.join(basedir, include)
                        immediate_includes[include_option].add((abs_include, include))
                        if deep and abs_include not in skip_paths:
                            _, new_abs_includes = cls.get_config_includes(abs_include, skip_paths + [abs_include])
                            deep_includes[include_option] += new_abs_includes

        if deep:
            return immediate_includes, deep_includes['before'] + [filename] + deep_includes['after']
        else:
            return immediate_includes

    def _list_configs(self, ext=None):
        """ Returns list of existing configuration files, limited to given a extension.
            If ext == None, both *.conf and *.local files are listed.
            Configuration file paths in the list include basedir and extensions.

            Files are listed in the following order: 
              ['basedir/name.conf', 'basedir/name.d/*.conf', 'basedir/name.local', 'basedir/name.d/*.local'].
            If name is a glob with '*' as file name, then only configs matching this glob under basedir are listed.

            This behavior (at least the first part) is how fail2ban reads configuration files.
        """
        if ext not in (None, 'conf', 'local'):
            raise ValueError("Invalid extension to list configuration files: %s" % ext)

        if ext is None:
            ext = ('conf', 'local')
        else:
            ext = (ext,)

        def gen_abs_config_path(basedir, basename, extension):
            return os.path.join(basedir, '%s.%s' % (basename, extension))

        log.debug("  Searching for %s configs representing '%s' under '%s'", ext, self._configs_name, self._basedir)

        config_files = []
        if os.path.basename(self._configs_name) == '*':
            for extension in ext:
                config_files += sorted(glob(gen_abs_config_path(self._basedir, self._configs_name, extension)))
        else:
            children_glob = os.path.join(self._configs_name + '.d', '*')
            for extension in ext:
                config_files.append(gen_abs_config_path(self._basedir, self._configs_name, extension))
                config_files += sorted(glob(gen_abs_config_path(self._basedir, children_glob, extension)))

        config_files = filter(os.path.exists, config_files)
        if config_files:
            log.debug("   Found following configs: %s", ', '.join(config_files))
        return config_files

    def _create_presentation(self, with_includes, only_ext):
        """ Creates new config presentation, reading files with specified extensions from 'basedir/name.ext',
            'basedir/name.d/anything.ext', properly reading includes as well, if requested.

            Result is either ConfigParser.RawConfigParser or ConfigParser.SafeConfigParser instance with 
            additional members:

            files - list of actually parsed configuration files, relative to basedir
            includes - dictionary with 'before' and 'after' keys, values are sets of immediately included
                       configuration files relative to basedir (or absolute paths); some files from these
                       sets may not exists or be valid configuration files
            save(filename=None) - a method available only when only_ext == 'local'; saves all local changes
                                  to name.local file (or filename if specified)
        """
        def merge_includes(a, b):
            """ Merge immediate_includes entry b into set of includes a relative to self._basedir. """
            for key in a:
                a[key] |= set(map(lambda (abs_include, include): strip_basedir(self._basedir, abs_include), b[key]))
            return a

        def save_local_config(self, filename=None):
            """ Save RawConfigParser instance to filename or default location (name.local). 
                Trying to save outside of initial basedir will trigger a warning, and file will not be saved.
                This is done mostly as a security precaution and to avoid causing potential havoc.
                This function is intended to be attached to a RawConfigParser instance as a method.
            """
            if not filename:
                filename = self._save_filename

            if os.path.commonprefix([filename, self._basedir]) != self._basedir:
                log.warning("Changes to configuration file '%s' will not be written since it is not under '%s'. "
                            "Please update it manually.", filename, self._basedir)
                return False

            if True:
                with open(filename, 'w') as output:
                    self.write(output)
                    log.debug("Successfully written %s", filename)
            else:   # Fake mode for testing
                log.debug("==== %s ====", filename)
                self.write(sys.stdout)
            return True

        root_config_files = self._list_configs(only_ext)

        immediate_includes = {}
        if with_includes:
            all_config_files = []
            for config in root_config_files:
                immediate_includes[config], deep_includes = self.get_config_includes(config)
                all_config_files += deep_includes
            config_parser_cls = ConfigParser.SafeConfigParser
        else:
            for config in root_config_files:
                immediate_includes[config] = self.get_config_includes(config, deep=False)
            all_config_files = root_config_files
            config_parser_cls = ConfigParser.RawConfigParser

        parser = config_parser_cls()
        parsed_configs = parser.read(all_config_files)

        # list of actually parsed configuration files, relative to basedir
        parser.files = map(functools.partial(strip_basedir, self._basedir), parsed_configs)
        # dict with two sets of immediately included configuration files, relative to basedir, some may not exist
        parser.includes = reduce(merge_includes, immediate_includes.itervalues(), {'before': set(), 'after': set()})
        # save() method for local configs
        if only_ext == 'local' and not with_includes:
            parser._basedir = self._basedir
            parser._save_filename = os.path.join(self._basedir, '%s.%s' % (self._configs_name, only_ext))
            parser.save = types.MethodType(save_local_config, parser)

        return parser

    def get_presentation(self, with_includes=True, only_ext=None):
        """ Get configuration file presentation. 
            By default presentation will include both .conf and .local files, and their deep includes.
        """
        if only_ext not in (None, 'conf', 'local'):
            raise ValueError("Invalid extension restriction argument: %s" % only_ext)

        presentation_hash = 'ext:%s includes:%s' % (only_ext, bool(with_includes))
        if presentation_hash not in self._presentations:
            self._presentations[presentation_hash] = self._create_presentation(with_includes, only_ext)
        return self._presentations[presentation_hash]

    def base(self):
        """ Get configuration presentation consisting of only base *.conf files, with their includes. 
            This will be an instance of ConfigParser.SafeConfigParser with .files and .includes additional members.
        """
        return self.get_presentation(only_ext='conf')

    def local(self):
        """ Get configuration presentation consisting of only customization *.local files, without includes.
            This will be an instance of ConfigParser.RawConfigParser with .file, .includes members and .save() method.
            Only this presentation should be considered writable or modifiable.
        """
        return self.get_presentation(with_includes=False, only_ext='local')

    def full(self, with_includes=True):
        """ Get configuration presentation consisting of the entire merged configuration, including 
            configuration files under name.d/ and included via [INCLUDES] section.
            Values of configuration options will be interpolated by default.
            This will be an instance of ConfigParser.SafeConfigParser with .files and .includes additional members.
        """
        return self.get_presentation(with_includes=with_includes)

    def exists(self):
        """ Returns True if the configuration is at all parsable (i.e. present at least partially). """
        return len(self.full().files) > 0

    def _enumerate_dependencies(self):
        """ Generates (relative config path w/o extension, config class, is in use) tuples for each config 
            this one depends on. Configs may repeat, with different 'is in use' flag values, but if there are
            tuples with 'is in use' == True, then they must go before tuples with 'is in use' == False for
            a given config.
            May be overriden in subclasses. Default implementation just lists before and after INCLUDES.
        """
        includes = self.full().includes
        for include in includes['before'] | includes['after']:
            yield (os.path.splitext(strip_basedir(self.SUBDIR, include))[0], self.__class__, True)

    def _merge_dependencies(self, source_cfg, configs_name_mapping, only_from_local=True):
        """ Merge immediately included configs from source_cfg into INCLUDES section, mapping configuration
            file names via configs_name_mapping.
            If only_from_local == False, then includes from base config file will be merged in addition to local.
        """
        source = source_cfg.local() if only_from_local else source_cfg.full(with_includes=False)
        target = self.local()

        if reduce(operator.or_, source.includes.values()) and not target.has_section(self.INCLUDES_SECTION):
            target.add_section(self.INCLUDES_SECTION)

        for include_order in source.includes:
            include_lines = set()
            for config in source.includes[include_order]:
                include_lines.add(strip_basedir(self.SUBDIR, configs_name_mapping.get(config, config)))
            if include_lines:
                target.set(self.INCLUDES_SECTION, include_order, join_multiline_value(include_lines))

    def _merge_section_from(self, source, section,
                            old_diff_source=None, new_diff_source=None, diffable_multiline_options=[]):
        """ Merge all local modifications to options from a section in a given source to .local().
            If an option is listed in diffable_multiline_options and present in both source and target base 
            configurations (old_diff_source and new_diff_source respectively), then it will be merged 
            "intellectually" by applying difference between base configurations to the local configuration.
        """
        def get_raw_value(config_parser, section, option):
            if isinstance(config_parser, ConfigParser.ConfigParser):
                return config_parser.get(section, option, raw=True)
            else:   # isinstance(config_parser, ConfigParser.RawConfigParser)
                return config_parser.get(section, option)

        if not isinstance(source, ConfigParser.RawConfigParser):
            raise ValueError("RawConfigParser instance is expected as source config, got: %s" % type(source))

        if section != self.DEFAULT_SECTION and not source.has_section(section):
            log.debug("There is no section '%s' in the source configuration.", section)
            return

        target = self.local()
        if section != self.DEFAULT_SECTION and not target.has_section(section):
            target.add_section(section)

        # We assume that source.items() returns raw values
        for option, value in source.items(section):
            # Skip options that came from 'DEFAULT' to another one (which we're processing)
            # Unfortunately there is no real way to know that option was not really declared in a given section,
            # but it just happens that it has the same value as in 'DEFAULT' section, so this is kinda hack :(
            if section != self.DEFAULT_SECTION and option in source.defaults() and value == source.defaults()[option]:
                continue

            if (option in diffable_multiline_options and 
                old_diff_source and old_diff_source.has_option(section, option) and 
                new_diff_source and new_diff_source.has_option(section, option)):

                lines = split_multiline_value(value)
                old_lines = split_multiline_value(get_raw_value(old_diff_source, section, option))
                new_lines = split_multiline_value(get_raw_value(new_diff_source, section, option))
                lines = apply_lines_diff(lines, *diff_lines(old_lines, new_lines))
                value = join_multiline_value(lines)

            target.set(section, option, value)

    def _merge_sections_from(self, source_cfg, source, old_diff_source=None, new_diff_source=None):
        """ Merge all relevant sections from another source. Should be overriden in subclasses. """
        self._merge_section_from(source, self.DEFAULT_SECTION)

    def _merge_from(self, source_cfg, processed_target_configs, configs_name_mapping, 
                    source_array, target_array, is_in_use=False):
        """ Merge configuration from source_cfg config file to this config file.

            processed_target_configs - list of already processed target config files (should be initially empty)
            configs_name_mapping - dictionary with config paths w/ extensions relative to basedir (initially empty)
            source_array - source configuration array (source_cfg should be spawned from it)
            target_array - target configuration array (self should be spawned from it)
            is_in_use - if True, this config file is referenced by something, so it should be present after merge

            This method updates processed_target_configs and configs_name_mapping, as well as sets self._merge_action
            which is later used by self._save_merge_changes() during the last stage of merge to commit it.
            This method also expects that _merge_sections_from(), _enumerate_dependencies(), and _merge_dependencies()
            are tailored to config types (that is, overriden in subclasses where appropriate).
        """
        def presentation_exists(presentation):
            return len(presentation.files) > 0

        if self in processed_target_configs:
            log.debug("Already processed: %s", self)
            return

        if not isinstance(source_cfg, ConfigPresentation):
            raise ValueError("ConfigPresentation instance is expected as source config, got: %s" % type(source_cfg))

        log.debug("Merging %s into %s", source_cfg, self)

        old_base_exists = presentation_exists(source_cfg.base())
        new_base_exists = presentation_exists(self.base())
        base_config_was_removed = old_base_exists and not new_base_exists

        if presentation_exists(source_cfg.local()) or (is_in_use and base_config_was_removed):

            if old_base_exists and new_base_exists:
                # Simple base config upgrade. Merge base differences into local config.
                self._merge_sections_from(source_cfg, source_cfg.local(), source_cfg.base(), self.base())
                self._merge_action = ('save', None)
                log.debug(" Merged '%s' - base config upgrade", self._configs_name)
            elif not old_base_exists and not new_base_exists:
                # Purely local customization. Just carry it over.
                self._merge_sections_from(source_cfg, source_cfg.local())
                self._merge_action = ('save', None)
                log.debug(" Merged '%s' - purely local customization", self._configs_name)
            elif old_base_exists and not new_base_exists:
                # Removal of base config on upgrade. Merge old base and local into new local config.
                self._merge_sections_from(source_cfg, source_cfg.full(with_includes=False))
                self._merge_action = ('save', None)
                log.debug(" Merged '%s' - removal of base config", self._configs_name)
            elif not old_base_exists and new_base_exists:
                # New config after upgrade has name clash with existing local config. Rename local config.
                self._merge_sections_from(source_cfg, source_cfg.local())
                new_local_filename = '%s-%s.local' % (self._configs_name, random_string(4))
                configs_name_mapping[self._configs_name + '.local'] = new_local_filename
                configs_name_mapping[self._configs_name + '.conf' ] = new_local_filename
                self._merge_action = ('save-as', os.path.join(self._basedir, new_local_filename))
                log.debug(" Merged '%s' - new base config clashes with local customization (local renamed to %s)", 
                          self._configs_name, new_local_filename)
        else:
            log.debug(" Skipped merge of '%s', as source isn't customized or in use", self._configs_name)
            self._merge_action = ('skip', None)

        processed_target_configs.append(self)

        for required_config_path, required_config_class, req_is_used in source_cfg._enumerate_dependencies():
            source_req_cfg = source_array.config(required_config_path, required_config_class)
            target_req_cfg = target_array.config(required_config_path, required_config_class)
            target_req_cfg._merge_from(source_req_cfg, processed_target_configs, configs_name_mapping, 
                                       source_array, target_array, is_in_use and req_is_used)

        # Observation: base configs will not include something that will require renaming
        if presentation_exists(source_cfg.local()) or (is_in_use and base_config_was_removed):
            self._merge_dependencies(source_cfg, configs_name_mapping, only_from_local=not base_config_was_removed)

    def _save_merge_changes(self):
        """ Save changes performed during self._merge_from() call to file system. """
        save_action, save_arg = self._merge_action
        if save_action == 'save':
            self.local().save()
        elif save_action == 'save-as':
            self.local().save(save_arg)
        elif save_action == 'skip':
            pass
        else:
            log.warning("Unknown merge save action: %s", save_action)


class FilterConfig(ConfigPresentation):
    SUBDIR = 'filter.d'

    ALL_OPTIONS = ('failregex', 'ignoreregex')

    def __init__(self, name, basedir):
        super(FilterConfig, self).__init__(os.path.join(self.SUBDIR, name), basedir)
        self._filter_name = name

    def is_complete(self):
        """ Check whether this config is a complete filter, rather than a part of another filter or filters. """
        try:
            return (self.exists() and 
                    self.full().has_section(self.DEFINITION_SECTION) and 
                    self.full().get(self.DEFINITION_SECTION, 'failregex') and
                    self.full().get(self.DEFINITION_SECTION, 'ignoreregex') is not None)
        except (ConfigParser.NoOptionError, ConfigParser.InterpolationError) as ex:
            log.debug("Value interpolation in %s (%s) failed: %s", 
                      self._configs_name, ', '.join(self.full().files), ex)
            return False

    def _merge_sections_from(self, source_cfg, source, old=None, new=None):
        super(FilterConfig, self)._merge_sections_from(source_cfg, source, old, new)
        self._merge_section_from(source, self.DEFINITION_SECTION, old, new, ('failregex', 'ignoreregex'))


class ActionConfig(ConfigPresentation):
    SUBDIR = 'action.d'

    ACTION_OPTIONS = ('actionstart', 'actionstop', 'actioncheck', 'actionban', 'actionunban')
    ALL_OPTIONS = ACTION_OPTIONS

    def __init__(self, name, basedir):
        super(ActionConfig, self).__init__(os.path.join(self.SUBDIR, name), basedir)
        self._action_name = name

    def is_complete(self):
        """ Check whether this config is a complete action, rather than a part of another action or actions. """
        try:
            return (self.exists() and
                    self.full().has_section(self.DEFINITION_SECTION) and # Init section is not obligatory
                    all(self.full().get(self.DEFINITION_SECTION, option) is not None 
                        for option in self.ACTION_OPTIONS))
        except (ConfigParser.NoOptionError, ConfigParser.InterpolationError) as ex:
            log.debug("Value interpolation in %s (%s) failed: %s", 
                      self._configs_name, ', '.join(self.full().files), ex)
            return False

    def default_parameters(self):
        """ Returns a list of (name, default value) pairs for action parameters. Valid only for complete actions. """
        if self.full().has_section(self.INIT_SECTION):
            return self.full().items(self.INIT_SECTION)
        else:
            return []

    def _merge_sections_from(self, source_cfg, source, old=None, new=None):
        super(ActionConfig, self)._merge_sections_from(source_cfg, source, old, new)
        self._merge_section_from(source, self.INIT_SECTION)
        self._merge_section_from(source, self.DEFINITION_SECTION)


class JailsConfig(ConfigPresentation):
    class JailConfig(object):
        BOOLEAN_OPTIONS = ('enabled',)
        INTEGER_OPTIONS = ('maxretry', 'findtime', 'bantime')
        MULTILINE_OPTIONS = ('logpath', 'action', 'failregex', 'ignoreregex')
        STRING_OPTIONS = ('backend', 'usedns', 'ignoreip', 'filter', 'ignorecommand')
        ALL_OPTIONS = BOOLEAN_OPTIONS + INTEGER_OPTIONS + MULTILINE_OPTIONS + STRING_OPTIONS
        ACTION_SPLIT_RE = re.compile("^((?:\w|-|_|\.)+)(?:\[(.*)\])?$")

        def __init__(self, jails, name, presentation='full'):
            self._jails = jails
            self._jail_name = name
            self._presentation = presentation

            if presentation == 'full':
                self._jails_config = self._jails.full()
            elif presentation == 'local':
                self._jails_config = self._jails.local()
            elif presentation == 'base':
                self._jails_config = self._jails.base()
            else:
                raise ValueError("Invalid presentation value for JailConfig: %s" % presentation)

        def exists(self):
            """ Check whether this jail exists. """
            return self._jails_config.has_section(self._jail_name)

        def is_enabled(self):
            """ Check whether this jail is enabled. """
            try:
                return self.get('enabled')
            except (ConfigParser.Error, ValueError) as ex:
                log.debug("Error while checking '%s' jail 'enabled' option: %s", self._jail_name, ex)
                return False

        def is_valid(self):
            """ Check whether this jail is syntactically valid, capable of being enabled. 
                To check actual validity within configuration array, one would also need to check that 
                referenced filter and actions do exist and are complete. Checking logs existence might be 
                useful as well.
            """
            try:
                return bool(self.exists() and
                            self.get('logpath') and self.get('filter') and list(self.actions()))
            except (ConfigParser.Error, ValueError) as ex:
                log.debug("Jail '%s' is not valid: %s", self._jail_name, ex)
                return False

        def rename_to(self, new_name):
            """ Rename jail. This can be used only for local presentation. 
                Raises ValueError if rename is not possible.
            """
            old_name = self._jail_name
            config = self._jails_config

            if new_name == JailsConfig.DEFAULT_SECTION or config.has_section(new_name):
                raise ValueError("Cannot rename jail from '%s' to '%s': invalid or already occupied name" %
                                 (old_name, new_name))
            if not self.exists():
                raise ValueError("Connot rename jail from '%s' to '%s': jail doesn't exist" % 
                                 (old_name, new_name))

            config.add_section(new_name)
            for option, value in config.items(old_name):
                if option in config.defaults() and value == config.defaults()[option]:
                    continue
                config.set(new_name, option, value)
            config.remove_section(old_name)

            self._jail_name = new_name

        def get(self, option):
            """ Get jail option value with proper type: boolean, int, string or list of strings. 
                Raises ConfigParser.Error descendants or ValueError on errors.
            """
            if option in self.BOOLEAN_OPTIONS:
                return self._jails_config.getboolean(self._jail_name, option)
            elif option in self.INTEGER_OPTIONS:
                return self._jails_config.getint(self._jail_name, option)
            elif option in self.MULTILINE_OPTIONS:
                return split_multiline_value(self._jails_config.get(self._jail_name, option))
            elif option in self.STRING_OPTIONS:
                return self._jails_config.get(self._jail_name, option)
            else:
                log.warning("Getting unknown jail option: %s", option)
                return self._jails_config.get(self._jail_name, option)

        def set(self, option, value):
            """ Set jail option value. This can be used only for local presentation. """
            self._jails_config.set(self._jail_name, option, str(value))

        @staticmethod
        def _split_action_parameters(action_opts_str):
            """ Returns generator that produces [name, value] pairs given action parameters string. 
                Invalid parameters w/o values are ignored with a warning.

                >>> list(JailsConfig.JailConfig._split_action_parameters('name=, port="ssh,25",protocol = tcp'))
                [['name', ''], ['port', 'ssh,25'], ['protocol', 'tcp']]
                >>> JailsConfig.JailConfig._split_action_parameters(None) # doctest: +ELLIPSIS
                <generator object ...>
                >>> [list(JailsConfig.JailConfig._split_action_parameters('  ')) for l in (None, '', '   ')]
                [[], [], []]
                >>> list(JailsConfig.JailConfig._split_action_parameters('ignored, empty= , '))
                [['empty', '']]
            """
            if not action_opts_str or not action_opts_str.strip():
                return

            lex = shlex(action_opts_str, posix=True)
            lex.commenters = ''
            lex.whitespace = ','
            lex.wordchars += '= \t'
            lex.whitespace_split = True
            for token in lex:
                if '=' not in token:
                    log.warning("Parameter '%s' in action parameters list '%s' doesn't have a value. Ignored.",
                                token.strip(), action_opts_str)
                    continue
                yield map(str.strip, token.split('=', 1))

        _ACTION_PARAM_VALUES_SAFECHARS = frozenset(string.ascii_letters + string.digits + '@%_-+:./')

        @staticmethod
        def _quote_action_param_value(string):
            """ Quote parameter value for action parameters string, if required.
                This method is loosely similar to standard pipes.quote(), except it also quotes commas and assignment.

                >>> print(JailsConfig.JailConfig._quote_action_param_value('space separated'))
                'space separated'
                >>> print(JailsConfig.JailConfig._quote_action_param_value('single-word'))
                single-word
                >>> print(JailsConfig.JailConfig._quote_action_param_value("quot'ed"))
                'quot'"'"'ed'
                >>> JailsConfig.JailConfig._quote_action_param_value('')
                ''
            """
            for char in string:
                if char not in JailsConfig.JailConfig._ACTION_PARAM_VALUES_SAFECHARS:
                    break
            else:
                return string if string else ''
            return "'" + string.replace("'", "'\"'\"'") + "'"

        @staticmethod
        def _assemble_action_parameters(action_opts):
            """ Returns a string representing given action parameters (list of [name, value] pairs). 
                This is a counterpart to _split_action_parameters() (except it adds square brakets).

                >>> actions = [['p1', "un"], ['p2', 'space sep'], ['p3', 'comma,sep=assign'], ['p4', ''], ['p5', "'"]]
                >>> C = JailsConfig.JailConfig
                >>> print(C._assemble_action_parameters(actions))
                [p1=un, p2='space sep', p3='comma,sep=assign', p4=, p5=''"'"'']
                >>> list(C._split_action_parameters(C._assemble_action_parameters(actions).strip('[]'))) == actions
                True
                >>> C._assemble_action_parameters([])
                ''
            """
            quote = JailsConfig.JailConfig._quote_action_param_value
            action_opts_str = ', '.join(map(lambda (name, value): '%s=%s' % (name, quote(value)), action_opts))
            return '[%s]' % action_opts_str if action_opts_str else ''

        def actions(self):
            """ Returns generator that produces (action name, [[param name, param value], ...]) tuples, enumerating
                all actions for this jail.
                Raises ValueError on parsing errors.
            """
            for action_line in self.get('action'):
                try:
                    match = self.ACTION_SPLIT_RE.match(action_line)
                    if match:
                        action_name, action_opts_str = match.groups()
                    else:
                        raise ValueError("Unable to split action definition into action and parameters")

                    yield action_name, list(self._split_action_parameters(action_opts_str))
                except Exception as ex:
                    raise ValueError("Invalid action '%s' in jail '%s': %s" % (action_line, self._jail_name, ex))

        def set_actions(self, actions):
            """ Replace all current actions with the given ones (a list of the same values actions() generates) 
                for this jail.
            """
            action_lines = []
            for action_name, params in actions:
                action_lines.append(action_name + self._assemble_action_parameters(params))
            self.set('action', join_multiline_value(action_lines))

        def filter(self):
            """ Returns filter name for this jail. """
            return self.get('filter')

        def set_filter(self, filter_name):
            """ Replace current filter name with a new one for this jail. """
            self.set('filter', filter_name)

    def __init__(self, name, basedir):
        super(JailsConfig, self).__init__('jail', basedir)

    def jail(self, name, presentation='full'):
        """ Returns JailsConfig.JailConfig object for given jail name with a given presentation (full by default). """
        return JailsConfig.JailConfig(self, name, presentation)

    def list(self, enabled=None, valid=None):
        """ List names of those jails that match specified properties. By default lists all jails.
            Full presentation is used to generate the list.
        """
        jails = self.full().sections()
        if enabled is not None:
            jails = filter(lambda name: self.jail(name).is_enabled() == enabled, jails)
        if valid is not None:
            jails = filter(lambda name: self.jail(name).is_valid() == valid, jails)
        return sorted(jails)

    def _enumerate_dependencies(self):
        def cmp_2nd(expected):
            def predicate(arg):
                return arg[1] == expected
            return predicate

        super(JailsConfig, self)._enumerate_dependencies()

        valid_jails = self.list(valid=True)
        jails_usage = map(lambda name: (name, self.jail(name, 'local').exists()), valid_jails)

        for jail_name, is_used in filter(cmp_2nd(True), jails_usage) + filter(cmp_2nd(False), jails_usage):
            jail = self.jail(jail_name)
            yield (jail.filter(), FilterConfig, is_used)
            for action_name, action_params in jail.actions():
                yield (action_name, ActionConfig, is_used)

    def _merge_dependencies(self, source_cfg, configs_name_mapping, only_from_local=True):
        def map_config_name(subdir, name):
            """ Maps filter or action name via configs_name_mapping to a new name. """
            path = os.path.join(subdir, '%s.local' % name)
            return os.path.splitext(os.path.basename(configs_name_mapping.get(path, path)))[0] if name else name

        super(JailsConfig, self)._merge_dependencies(source_cfg, configs_name_mapping, only_from_local)

        # This assumes that 'filter' and 'action' fields might have been already updated, so local already has changes
        for jail_name in self.local().sections():
            full_jail = self.jail(jail_name)            # May not exist, if jail was renamed
            local_jail = self.jail(jail_name, 'local')

            # Check and update filter
            cur_filter = None
            try:
                if full_jail.exists():
                    cur_filter = full_jail.filter()
                cur_filter = local_jail.filter()
            except ConfigParser.NoOptionError:
                pass

            new_filter = map_config_name(FilterConfig.SUBDIR, cur_filter)
            if new_filter != cur_filter and new_filter:
                log.debug("In jail '%s' filter was changed from '%s' to '%s'", jail_name, cur_filter, new_filter)
                local_jail.set_filter(new_filter)

            # Check and update actions
            cur_actions = []
            try:
                if full_jail.exists():
                    cur_actions = list(full_jail.actions())
                cur_actions = list(local_jail.actions())
            except (ConfigParser.NoOptionError, ValueError):
                pass

            new_actions = map(lambda (name, params): (map_config_name(ActionConfig.SUBDIR, name), params), 
                              cur_actions)
            if new_actions != cur_actions and new_actions:
                log.debug("In jail '%s' actions were changed from '%s' to '%s'", jail_name, cur_actions, new_actions)
                local_jail.set_actions(new_actions)

    def _merge_sections_from(self, source_cfg, source, old=None, new=None):
        super(JailsConfig, self)._merge_sections_from(source_cfg, source, old, new)

        discarded_unmodified_jails = []

        for jail_name in source_cfg.list(valid=True):
            old_base_exists = source_cfg.jail(jail_name, 'base').exists()
            new_base_exists = self.jail(jail_name, 'base').exists()

            if not new_base_exists and not source_cfg.jail(jail_name, 'local').exists():
                discarded_unmodified_jails.append(jail_name)
                continue

            if old_base_exists and not new_base_exists:
                jail_source = source_cfg.full(with_includes=False)
            else:
                jail_source = source

            self._merge_section_from(jail_source, jail_name, old, new, ('logpath', 'failregex', 'ignoreregex'))

            new_jail_name = jail_name
            if not old_base_exists and new_base_exists:
                new_jail_name = '%s-%s' % (jail_name, random_string(4))
                self.jail(jail_name, 'local').rename_to(new_jail_name)

            self.__merge_missing_options_from(source_cfg, jail_name, new_jail_name, self.JailConfig.ALL_OPTIONS)

        invalid_jails = source_cfg.list(valid=False)
        if invalid_jails:
            log.warning("Following jails were discarded during configuration update as they were misconfigured: %s",
                        ', '.join(invalid_jails))
        if discarded_unmodified_jails:
            log.warning("Following jails were discarded during configuration update as they didn't have local "
                        "customizations and aren't present in base configuration anymore: %s",
                        ', '.join(discarded_unmodified_jails))

    __INTERPOLATION_KEY_RE = re.compile(r"%\(([^)]*)\)s")

    def __merge_missing_options_from(self, source_cfg, section, target_section, options):
        """ Merge interpolated values of given options missing in target_section of this configuration 
            (base or local) from source_cfg section or [DEFAULT].
        """
        # To anyone who would want to simply move this method to ConfigPresentation to call it for filters and 
        # actions as well: you will probably break something. It was designed for jails config only, since
        # it (in practice) doesn't include other configs and all valid jails need to have a predetermined set
        # of options.
        #
        # Normally one would need to implement per-option merge that takes interpolation (and hence options
        # interdependencies) into account instead of this method. Unfortunately I had no time to properly do so 
        # and the approach I used results in much less "configuration noise" after the merge, yielding minimum
        # number of additional options only where they are actually required, while not affecting other jails.
        #
        # This method is primarily required for smooth upgrade from Debian system fail2ban configuration.
        source = source_cfg.full()
        target = self.local()
        base = self.base()

        if target_section == self.DEFAULT_SECTION or not source.has_section(section):
            return

        if not target.has_section(target_section):
            target.add_section(target_section)

        for option in options:
            if ((target.has_option(target_section, option) and 
                 not self.__INTERPOLATION_KEY_RE.match(target.get(target_section, option))) or 
                base.has_option(target_section, option) or 
                base.has_option(self.DEFAULT_SECTION, option)):
                continue

            for source_section in (section, self.DEFAULT_SECTION):
                if source.has_option(source_section, option):
                    interpolated_value = source.get(source_section, option)
                    target.set(target_section, option, interpolated_value)
                    log.debug(" Added missing option '%s' to '%s' jail "
                              "(value '%s' was interpolated from source configuration section '%s')", 
                              option, target_section, interpolated_value, source_section)
                    break


class GeneralConfig(ConfigPresentation):
    ALL_OPTIONS = ('loglevel', 'logtarget', 'socket', 'pidfile')

    def __init__(self, name, basedir):
        super(GeneralConfig, self).__init__('fail2ban', basedir)

    def _merge_sections_from(self, source_cfg, source, old=None, new=None):
        super(GeneralConfig, self)._merge_sections_from(source_cfg, source, old, new)
        self._merge_section_from(source, self.DEFINITION_SECTION)


class ConfigurationArray(object):
    DEFAULT_BASEDIR = '/etc/fail2ban'
    JAILS_ROOT = 'jail'
    GENERAL_ROOT = 'fail2ban'

    def __init__(self, basedir=None):
        self.set_basedir(basedir)
        self._configs = {}

    def set_basedir(self, basedir):
        """ Set configuration array base directory. """
        self._basedir = basedir or self.DEFAULT_BASEDIR

    def config(self, name, cls=ConfigPresentation):
        """ Get configuration object for a given config (identified by relative name or glob w/o extension). """
        config_hash = cls.__module__ + '.' + cls.__name__ + '::' + name
        if config_hash not in self._configs:
            self._configs[config_hash] = cls(name, self._basedir)
        return self._configs[config_hash]

    def filter(self, name):
        return self.config(name, FilterConfig)

    def action(self, name):
        return self.config(name, ActionConfig)

    def jails(self):
        return self.config('jail', JailsConfig)

    def fail2ban(self):
        return self.config('fail2ban', GeneralConfig)

    def _glob_configs(self, subdir, extensions=('conf', 'local')):
        """ Returns sorted list of configuration files' basenames w/o extensions in a given subdirectory 
            that all have given extensions.
        """
        if not isinstance(extensions, (list, tuple)):
            extensions = [extensions]

        paths = set()
        for ext in extensions:
            paths.update(map(lambda path: os.path.splitext(os.path.basename(path))[0], 
                             glob(os.path.join(self._basedir, subdir, '*.%s' % ext))))
        return sorted(paths)

    def list_filters(self):
        return self._glob_configs('filter.d')

    def list_actions(self):
        return self._glob_configs('action.d')

    def merge_from(self, source_array):
        """ Merge local changes from source_array into this configuration array. """
        processed_configs = []
        configs_name_mapping = {}
        args = (processed_configs, configs_name_mapping, source_array, self)

        log.info("Merging 'fail2ban' configuration...")
        self.fail2ban()._merge_from(source_array.fail2ban(), *args, is_in_use=True)

        log.info("Merging 'jail' configuration...")
        self.jails()._merge_from(source_array.jails(), *args, is_in_use=True)

        log.info("Merging unprocessed 'filter.d/' configuration...")
        for filter_name in source_array.list_filters():
            self.filter(filter_name)._merge_from(source_array.filter(filter_name), *args)

        log.info("Merging unprocessed 'action.d/' configuration...")
        for action_name in source_array.list_actions():
            self.action(action_name)._merge_from(source_array.action(action_name), *args)

        log.info("Saving merge changes to file system...")

        for config in processed_configs:
            config._save_merge_changes()

        log.info("Configuration merge completed")

    def errors(self, check_only_in_use=True):
        """ Generates list of found configuration errors (error messages). 
            If check_only_in_use == True then only entities in use are checked (enabled jails and their 
            dependencies), otherwise everything is checked.
        """
        def filter_errors(name, jail_name):
            if not self.filter(name).exists():
                yield "Filter '%s' used in jail '%s' doesn't exist" % (name, jail_name)
            elif not self.filter(name).is_complete():
                yield "Filter '%s' used in jail '%s' is incomplete" % (name, jail_name)

        def action_errors(name, jail_name):
            if not self.action(name).exists():
                yield "Action '%s' used in jail '%s' doesn't exist" % (name, jail_name)
            elif not self.action(name).is_complete():
                yield "Action '%s' used in jail '%s' is incomplete" % (name, jail_name)

        for jail_name in self.jails().list(enabled=check_only_in_use or None):
            jail = self.jails().jail(jail_name)
            if not jail.is_valid():
                yield "%s jail '%s' is not valid" % ("Enabled" if jail.is_enabled() else "Disabled", jail_name)
                continue

            for error in filter_errors(jail.filter(), jail_name):
                yield error

            for action_name, params in jail.actions():
                for error in action_errors(action_name, jail_name):
                    yield error


def split_multiline_value(value):
    r""" Split multiline value into lines.

        >>> split_multiline_value('line1\n\r\n line2\nline3')
        ['line1', ' line2', 'line3']
        >>> split_multiline_value('')
        []
    """
    return filter(None, value.splitlines())

def join_multiline_value(lines):
    r""" Joins multiple lines into a single multiline value.

        >>> join_multiline_value(['line1', ' line2', 'line3'])
        'line1\n line2\nline3'
        >>> join_multiline_value([])
        ''
    """
    return '\n'.join(lines)

def diff_lines(old_lines, new_lines):
    """ Calculate ordered difference between two lists of lines. Returns (add, remove) sets.

        >>> diff_lines(['old', 'com', 'old'], ['com', 'new', 'new2', 'new']) == (set(['new', 'new2']), set(['old']))
        True
        >>> diff_lines([], [])
        (set([]), set([]))
    """
    add    = set(new_lines).difference(old_lines)
    remove = set(old_lines).difference(new_lines)
    return add, remove

def apply_lines_diff(lines, add_lines, remove_lines):
    """ Apply diff produced by diff_lines() (normally add and remove sets of lines) to a list of lines.

        >>> apply_lines_diff(['com', 'add', 'old'], *diff_lines(['old', 'rm', 'com'], ['new2', 'new', 'com']))
        ['com', 'add', 'new', 'new2']
        >>> apply_lines_diff([], [], [])
        []
    """
    return [line for line in lines if line not in remove_lines] + sorted(add_lines)

def strip_basedir(basedir, path):
    """ Strip a leading basedir from a given file path.
        Otherwise leave it as is, provided it is an absolute path.

        >>> strip_basedir('relative/path', 'relative/path/to/file')
        'to/file'
        >>> strip_basedir('/absolute/path', '/absolute/path/to/file')
        'to/file'
        >>> strip_basedir('/absolute/path', '/another/path/to/file')
        '/another/path/to/file'
        >>> strip_basedir('', './file')
        'file'
        >>> strip_basedir('relative/path', 'another/path/to/file')
        Traceback (most recent call last):
            ...
        AssertionError
    """
    if os.path.commonprefix([path, basedir]) == basedir:
        return os.path.relpath(path, basedir)
    else:
        assert os.path.isabs(path)
        return path

def random_string(length, charset=string.ascii_uppercase + string.digits):
    """ Returns a random string with a given length, each character is from the charset.

        >>> len(random_string(10))
        10
        >>> random_string(3, '0')
        '000'
        >>> random_string(2, 'ab') in ('aa', 'ab', 'ba', 'bb')
        True
    """
    return ''.join(random.choice(charset) for i in range(length))


if __name__ == '__main__':
    import doctest, sys, shutil

    doctest_rc = doctest.testmod().failed
    if doctest_rc:
        sys.exit(doctest_rc)

    def copy_overwrite_dir(src, dst):
        if os.path.isdir(dst):
            shutil.rmtree(dst)
        shutil.copytree(src, dst)

    # Prepare test data
    source = './test-data/fail2ban.src'
    vanilla_target = './test-data/fail2ban.tgt'
    target = './test-data/fail2ban.res'

    copy_overwrite_dir(vanilla_target, target)

    # Merge source into target
    src_array = ConfigurationArray(source)
    tgt_array = ConfigurationArray(target)
    tgt_array.merge_from(src_array)

    merge_errors = list(ConfigurationArray(target).errors(check_only_in_use=False))
    if merge_errors:
        log.error("Following configuration errors were found after merge (turn on debug logging to see details):")
        map(lambda msg: log.error(" %s", msg), merge_errors)
        sys.exit(1)

# vim: ts=4 sts=4 sw=4 et :
