# Copyright 1999-2014. Parallels IP Holdings GmbH. All Rights Reserved.
""" spammng - SpamAssassin Plesk backend management application """

import sys
import string
import subprocess
import os
import errno
import pwd
import shutil
import stat
import re

from optparse import OptionParser

import plesk_log
import plesk_config
import plesk_service
from os_detect import OS
from plesk_file_utils import ReplacementFile

import sa_layout


log = plesk_log.getLogger(__name__)

POPUSER = 'popuser'


class MultipleExceptions(Exception):
    """ Exception that represents an array of exceptions. """
    def __init__(self, exceptions):
        self.exceptions = exceptions

    def __str__(self):
        msg = 'Multiple errors occurred:\n'
        for ex in self.exceptions:
            msg += str(ex) + '\n'
        return msg

class MultipleCalledProcessErrors(MultipleExceptions):
    pass

class SpamAssassinServiceException(Exception):
    pass


_TRANSLATE_ADDR_SEP_TO_SPACES = string.maketrans(',', ' ')

def normalize_address_list(addrs):
    """ Convert mail address and pattern lists from command line to internal representation.

        Converts a list of entries like 'mailname, mailname mailname' into a plain list 
        of 'mailname' entries.
    """
    if not addrs:
        return []
    addrs_matrix = [ sublist.translate(_TRANSLATE_ADDR_SEP_TO_SPACES).split() for sublist in addrs ]
    return reduce(lambda x, y: x+y, addrs_matrix, [])

def lowercase_list(addrs):
    """ Convert all entries in a list to lowercase. """
    return map(lambda x: x.lower(), addrs or [])


def spam_mail_handlers_control(action, mailname=None, context=None):
    """ Invoke mail_handlers_control with appropriate arguments for 'spam' handler.
        Raises subprocess.CalledProcessError if execution failed.
    """
    if action not in ('add', 'remove', 'enable', 'disable'):
        raise ValueError("action value '%s' is invalid" % action)
    if action in ('add', 'remove') and not mailname:
        raise ValueError("mailname is required for '%s' action" % action)

    args = {'name': sa_layout.SPAM_HOOK_NAME,
            'type': 'recipient',
            'queue': 'before-local',
           }

    if mailname:
        args['mailname'] = mailname

    if action == 'add':
        if not context:
            raise ValueError("context is require for '%s' action" % action)
        args.update({'context': context,
                     'priority': sa_layout.SPAM_HOOK_PRIORITY,
                     'executable': sa_layout.SPAM_HOOK_BIN,
                     'enabled': None,
                     'dont-preserve-on-restore' : None,
                    })

    args_list = [os.path.join(sa_layout.ADMIN_SBIN_DIR, 'mail_handlers_control'), '--' + action]
    for opt, val in args.iteritems():
        args_list.append('--' + opt + (val and '=' + val or ''))

    log.debug("Calling mail_handlers_control with: %s", args_list)
    subprocess.check_call(args_list)

def mass_spam_mail_handlers_control(action, mailnames, context=None):
    """ Invokes mail_handlers_control for all given mailnames.
        May raise MultipleCalledProcessErrors only after all mailnames are processed.
        Note: this may be implemented more efficiently via mail_handlers_control
        mass operations support.
    """
    exceptions = []
    for mailname in mailnames:
        try:
            spam_mail_handlers_control(action, mailname, context)
        except subprocess.CalledProcessError, ex:
            exceptions.append(ex)
    if exceptions:
        raise MultipleCalledProcessErrors(exceptions)


class SpamAssassinService:
    """ This class provides methods to control SpamAssassin service. """
    def __init__(self, service_configuration):
        self._service_config = service_configuration
        self._is_enabled = None

    def enable(self):
        """ Enables SpamAssassin service. """
        self._service_config.enable_service()
        plesk_service.register(sa_layout.get_service_name())
        self._is_enabled = True

    def disable(self):
        """ Disables SpamAssassin service. Also stops it before disabling. """
        plesk_service.action(sa_layout.get_service_name(), 'stop')
        self._service_config.disable_service()
        plesk_service.deregister(sa_layout.get_service_name())
        self._is_enabled = False

    def is_enabled(self):
        """ Check whether service is enabled. May not be accurate accross all OSes. """
        if self._is_enabled is None:
            self._is_enabled = self._service_config.is_service_enabled()
        return self._is_enabled

    def action(self, act):
        """ Perform SpamAssassin service action (start, restart, etc.). If service is disabled 
            action may not be performed, but instead the service will be considered stopped.
        """
        if act not in ('start', 'stop', 'restart', 'condrestart', 'status'):
            raise ValueError("Unsupported SpamAssassin service action '%s'" % act)
        if not self.is_enabled():
            log.warning("Service is not enabled in configuration! Service %s will not be actually performed.", act)
            # Following line should really be: return act in ('stop', 'condrestart'); but BL is too buggy :'(
            return act != 'status'
        emulate = act in ('condrestart',)
        return plesk_service.action(sa_layout.get_service_name(), act, emulate)


class SpamAssassinServiceConfiguration:
    """ This class provides methods to edit SpamAssassin service configuration. """
    _comment_line_re = re.compile('^\s*#')
    _config_line_re = re.compile('^\s*(\S+?)\s*=\s*(.*)\n$')

    def __init__(self):
        pass

    @staticmethod
    def get_default_max_children():
        """ Detect default value for --max-children option. """
        return OS.is_vz_based() and 1 or 5

    def _build_config_line(self, option, new_value):
        if new_value is None:
            return None
        elif isinstance(new_value, (int, long)):
            return '%s=%d\n' % (option, new_value)
        else:
            return '%s="%s"\n' % (option, new_value)

    def _read_config(self, option=None):
        """ Read service configuration file and dictionary with lowercased options as keys
            or options value if option name is provided as argument or None if option was
            not encountered in the configuration file. Quotes are trimmed from values.
        """
        options = {}
        config = sa_layout.get_service_config_path()
        with open(config, 'r') as cfg:
            for line in cfg:
                if not line.strip() or self._comment_line_re.match(line):
                    pass
                else:
                    data = self._config_line_re.match(line)
                    if data:
                        opt = data.group(1).lower()
                        val = data.group(2).strip().strip('"').strip("'")
                        options[opt] = val
        if option is None:
            return options
        else:
            return options.get(option.lower(), None)

    def _update_config(self, **kwargs):
        """ Updates service configuration file replacing specified variable values. """
        config = sa_layout.get_service_config_path()
        with ReplacementFile(config, 0644, 0, 0) as new:
            with open(config, 'r') as old:
                for line in old:
                    if not line.strip() or self._comment_line_re.match(line):
                        pass
                    else:
                        data = self._config_line_re.match(line)
                        if data and data.group(1).lower() in kwargs.keys():
                            option = data.group(1)
                            line = self._build_config_line(option, kwargs[option.lower()])
                            del kwargs[option.lower()]
                    if line is not None:
                        new.write(line)
            for option, value in kwargs.iteritems():
                line = self._build_config_line(option, value)
                if line is not None:
                    new.write(line)

    def update(self, enable_user_configs, max_children=None):
        """ Update SpamAssassin service configuration (daemon start options). """
        if max_children is None:
            max_children = self.get_default_max_children()
        plesk_mailnames_d = plesk_config.get('PLESK_MAILNAMES_D')

        args = {'--username': POPUSER,
                '--daemonize': None,
                '--nouser-config': None,
                '--helper-home-dir': sa_layout.QMAIL_DIR,
                '--max-children': max_children,
               }
        if enable_user_configs:
            args.update({'--create-prefs': None,
                         '--virtual-config-dir': plesk_mailnames_d + '/%d/%l/.spamassassin',
                        })
        args_str = ' '.join( map(lambda (k, v): v is None and k or '%s=%s' % (k, v), args.iteritems()) )

        if OS.is_redhat_based():
            self._update_config(spamdoptions=args_str)
        elif OS.is_deb_based():
            self._update_config(options=args_str)
        elif OS.is_suse():
            self._update_config(spamd_args=args_str)
        else:
            raise RuntimeError("Unexpected OS distribution: %s" % OS.distname())

    def enable_service(self):
        """ Enables SpamAssassin service in configuration. """
        if OS.is_deb_based():
            self._update_config(enabled=1)

    def disable_service(self):
        """ Disables SpamAssassin service in configuration. Remember to stop it before disabling! """
        if OS.is_deb_based():
            self._update_config(enabled=0)

    def is_service_enabled(self):
        """ Check if SpamAssassin is enabled in configuration. """
        if OS.is_deb_based():
            enabled = self._read_config('enabled')
            try:
                return enabled is not None and int(enabled)
            except ValueError, ex:
                log.warning("Unexpected value '%s' for 'ENABLED' option in service config: %s", enabled, ex)
                return False
        return True


class SpamAssassinFilterConfiguration:
    """ This class provides methods to edit SpamAssassin server-wide and user 
        configuration files.

        Only a restricted set of configuration directives is processed, other
        directives will be left unchanged. The class operates on server-wide
        configuration file (local.cf) if mailname is not specified in constructor.
        Use update() method to actually make any scheduled changes to a file.
    """
    _overridable_settings = ('required_score', 'rewrite_header')
    _list_settings = ('whitelist_from', 'blacklist_from', 'unwhitelist_from', 'unblacklist_from')

    def __init__(self, mailname=None):
        self.mailname = mailname
        self._overrides = {}
        self._list_additions = {}
        self._list_removals = {}
        self._reset = False

    def is_server_config(self):
        """ Check if this filter configuration instance is for server-wide config. """
        return not self.mailname

    def _user_config_dir(self):
        """ Returns user SpamAssassin configuration directory path.
            The directory will be created and proper permissions set if required.
            Applicable only if not is_server_config().
        """
        if self.is_server_config():
            raise ValueError("Can't deduce user config path since mailname was not specified")

        mail_localpart, mail_domain = self.mailname.split('@')
        plesk_mailnames_d = plesk_config.get('PLESK_MAILNAMES_D')
        user_config_d = os.path.join(plesk_mailnames_d, mail_domain, mail_localpart, '.spamassassin')

        # Ensure that configuration directory exists and has correct access rights
        try:
            os.makedirs(user_config_d, 0755)
        except OSError, ex:
            if ex.errno != errno.EEXIST:
                raise
        else:
            popuser = pwd.getpwnam(POPUSER)
            os.chown(user_config_d, popuser.pw_uid, popuser.pw_gid)

        return user_config_d

    def _config_path(self):
        """ Returns path to SpamAssassin configuration file. """
        if self.is_server_config():
            return sa_layout.SA_SERVER_CONFIG
        else:
            return os.path.join(self._user_config_dir(), 'user_prefs')

    def reset_settings(self):
        self._reset = True

    def override_setting(self, param, value):
        if param not in self._overridable_settings:
            raise ValueError("Unexpected SpamAssassin setting to replace: '%s'" % param)
        self._overrides[param] = value

    def add_to_list_setting(self, param, values):
        if param not in self._list_settings:
            raise ValueError("Unexpected SpamAssassin setting to add to: '%s'" % param)
        self._list_additions[param] = self._list_additions.get(param, []) + values

    def remove_from_list_setting(self, param, values):
        if param not in self._list_settings:
            raise ValueError("Unexpected SpamAssassin setting to remove from: '%s'" % param)
        self._list_removals[param] = self._list_removals.get(param, []) + values

    def _filter_config_line(self, option, value):
        """ Process configuration file line and return new value for the option in this line.
            Returns None to indicate that the line should be removed.
        """
        if self._reset and option in self._list_settings + self._overridable_settings:
            return None
        if option in self._list_settings:
            values = value.split()
            values = list( set(values) - 
                            set(self._list_removals.get(option, [])) - 
                            set(self._list_additions.get(option, [])) )
            return values and ' '.join(values) or None
        elif option in self._overrides.keys():
            return None
        return value

    def update(self):
        """ Update selected configuration file with changes scheduled previously via
            override_setting(), add_to_list_setting() and remove_from_list_setting()
            methods. Note that first all deletions are performed and then new and changed
            values are appended to configuration file.

            Implementation note: technically current implementation is not entirely correct
            for order-sesitive directives, namely 'unwhitelist_from' and 'unblacklist_from'.
            E.g. if a given address mask is both in whitelist and unwhitelist for a mailname,
            then it is possible to have it first appear in 'unwhitelist_from' and then in
            'whitelist_from' directives after several spammng calls. But having same address
            mask in both lists on the same level (that is user prefs level) is insane and 
            useless. Also, IMO, un*list support in Plesk is broken and nobody uses it anyway.
        """
        comment_line_re = re.compile('^\s*#')
        config_line_re = re.compile('^\s*(\S+)\s+(.*)\n$')

        config = self._config_path()
        popuser = pwd.getpwnam(POPUSER)
        with ReplacementFile(config, 0600, popuser.pw_uid, popuser.pw_gid) as new:
            firstline = True
            try:
                with open(config, 'r') as old:
                    for line in old:
                        if firstline and line.strip() != sa_layout.AUTOGENERATED_CONFIG_NOTICE.split('\n')[0].strip():
                            new.write(sa_layout.AUTOGENERATED_CONFIG_NOTICE)
                        firstline = False

                        if not line.strip() or comment_line_re.match(line):
                            pass
                        else:
                            data = config_line_re.match(line)
                            if data:
                                option, value = data.group(1), data.group(2)
                                value = self._filter_config_line(option, value)
                                if value is None:
                                    continue
                                line = "%s %s\n" % (option, value)
                        new.write(line)
            except IOError, ex:
                if ex.errno != errno.ENOENT:
                    raise
                log.debug("Source config '%s' doesn't exist - a new one will be created", config)

            if firstline:
                new.write(sa_layout.AUTOGENERATED_CONFIG_NOTICE)
            for option, value in self._overrides.iteritems():
                if value is not None:
                    new.write("%s %s\n" % (option, value))
            for option, values in self._list_additions.iteritems():
                if values:
                    new.write("%s %s\n" % (option, ' '.join(values)))


class SpamAssassinController:
    """ Top-level SpamAssassin management interface. """
    def __init__(self):
        self._service_configuration = SpamAssassinServiceConfiguration()
        self._service = SpamAssassinService(self._service_configuration)

    def enable(self):
        """ Enables SpamAssassin for all already registered mailnames. Registers service. """
        spam_mail_handlers_control('enable')
        self._service.enable()

    def disable(self):
        """ Disables SpamAssassin for all already registered mailnames. 
            Stops and deregisters service. 
        """
        spam_mail_handlers_control('disable')
        self._service.disable()

    def add_mailnames(self, mails, action):
        """ (Re)registers and enables mail handlers for all mailnames in the list. """
        if action not in ('text', 'delete', 'move'):
            raise ValueError("Mailname action value '%s' is invalid" % action)

        try:
            mass_spam_mail_handlers_control('remove', mails)
        except MultipleCalledProcessErrors, ex:
            log.error("Errors occured while removing some mail handlers: %s", ex)
        mass_spam_mail_handlers_control('add', mails, action)

    def remove_mailnames(self, mails):
        """ Deregisters mail handlers for all mailnames in the list. """
        mass_spam_mail_handlers_control('remove', mails)

    def enable_mailnames(self, mails):
        """ Enables mail handlers for all mailnames in the list. """
        mass_spam_mail_handlers_control('enable', mails)

    def disable_mailnames(self, mails):
        """ Disables mail handlers for all mailnames in the list. """
        mass_spam_mail_handlers_control('disable', mails)

    def service_action(self, action):
        """ Execute service action (start, stop, etc.). """
        if not self._service.action(action):
            raise SpamAssassinServiceException("Unable to %s SpamAssassin service" % action)
        return True

    def service_is_running(self):
        """ Return status of the service. True is running. """
        return self._service.action('status')

    def get_configuration(self, mailname=None):
        return SpamAssassinFilterConfiguration(mailname)

    def get_service_configuration(self):
        return self._service_configuration


class SpammngApp:
    def __init__(self):
        self.options = {}
        self.spamassassin = None

    def _execute_config_update(self):
        for mailname in self.options.mailname or [None]:
            config = self.spamassassin.get_configuration(mailname)

            if self.options.reset_config:
                config.reset_settings()
            if self.options.sensitivity is not None:
                config.override_setting('required_score', self.options.sensitivity)
            if self.options.modify_header is not None:
                config.override_setting('rewrite_header', 'subject ' + self.options.modify_header)
            # [[[cog
            #   import cog, itertools
            #   method_map = { 'add_to_': 'add_to_list_setting', 'del_from_': 'remove_from_list_setting' }
            #   for a, b, c in itertools.product(('add_to_', 'del_from_'), ('', 'un'), ('whitelist', 'blacklist')):
            #       cog.outl("""\
            #                   if self.options.%(opt)s:
            #                       config.%(method)s('%(setting)s', self.options.%(opt)s)""" %
            #               {'opt':     a + b + c,
            #                'method':  method_map[a],
            #                'setting': b + c + '_from',
            #               })
            # ]]]
            if self.options.add_to_whitelist:
                config.add_to_list_setting('whitelist_from', self.options.add_to_whitelist)
            if self.options.add_to_blacklist:
                config.add_to_list_setting('blacklist_from', self.options.add_to_blacklist)
            if self.options.add_to_unwhitelist:
                config.add_to_list_setting('unwhitelist_from', self.options.add_to_unwhitelist)
            if self.options.add_to_unblacklist:
                config.add_to_list_setting('unblacklist_from', self.options.add_to_unblacklist)
            if self.options.del_from_whitelist:
                config.remove_from_list_setting('whitelist_from', self.options.del_from_whitelist)
            if self.options.del_from_blacklist:
                config.remove_from_list_setting('blacklist_from', self.options.del_from_blacklist)
            if self.options.del_from_unwhitelist:
                config.remove_from_list_setting('unwhitelist_from', self.options.del_from_unwhitelist)
            if self.options.del_from_unblacklist:
                config.remove_from_list_setting('unblacklist_from', self.options.del_from_unblacklist)
            # [[[end]]] (checksum: af274fa7b9e7eec7c7c1699da052fa58)

            config.update()

    def _execute_service_update(self):
        config = self.spamassassin.get_service_configuration()
        config.update(self.options.enable_user_configs, self.options.max_children)

    def execute(self):
        self.spamassassin = SpamAssassinController()

        if self.options.enable:
            self.spamassassin.enable()

        if self.options.disable:
            self.spamassassin.disable()

        if self.options.enable_mailname:
            self.spamassassin.add_mailnames(self.options.mailname, self.options.action)

        if self.options.disable_mailname:
            self.spamassassin.remove_mailnames(self.options.mailname)

        if self.options.enable_checking:
            self.spamassassin.enable_mailnames(self.options.mailname)

        if self.options.disable_checking:
            self.spamassassin.disable_mailnames(self.options.mailname)

        if (self.options.sensitivity is not None or 
            self.options.modify_header is not None or
            self.options.reset_config or
            self.options.add_to_whitelist     or self.options.add_to_blacklist or
            self.options.add_to_unwhitelist   or self.options.add_to_unblacklist or
            self.options.del_from_whitelist   or self.options.del_from_blacklist or
            self.options.del_from_unwhitelist or self.options.del_from_unblacklist):

            self._execute_config_update()

        # Note that self.options.enable_server_configs is effectively unused and always on
        if ((self.options.enable_user_configs or self.options.max_children) and 
            self.options.service_action in ('update', 'start', 'restart', 'condrestart')):

            self._execute_service_update()

        if self.options.service_action in ('start', 'stop', 'restart', 'condrestart'):
            if self.spamassassin.service_action(self.options.service_action):
                print("Service %s was done successfully." % self.options.service_action)
        elif self.options.service_action in ('status',):
            if self.spamassassin.service_is_running():
                print("is running")
            else:
                print("is stopped")

    def check_options_sanity(self, options, error):
        if len(filter(None, [options.enable, options.disable, 
                             options.enable_mailname, options.disable_mailname, 
                             options.enable_checking, options.disable_checking])) > 1:

            error("the following options cannot be combined: --enable, --disable, "
                  "--enable-mailname, --disable-mailname, --enable-checking, --disable-checking")

        if ((options.enable_server_configs or options.enable_user_configs or options.max_children) and 
            options.service_action not in ('update', 'start', 'restart', 'condrestart')):

            error("any of --enable-server-configs, --enable-user-configs, or --max-children "
                  "options will have no effect unless service is updated (one of "
                  "--update, --start, --restart, or --condrestart is specified)")

    def preprocess_options(self, options):
        options.mailname = lowercase_list(normalize_address_list(options.mailname))
        # [[[cog
        #   import cog, itertools
        #   for a, b, c in itertools.product(('add_to_', 'del_from_'), ('', 'un'), ('whitelist', 'blacklist')):
        #       cog.outl('options.%(opt)s = lowercase_list(normalize_address_list(options.%(opt)s))' % { 'opt': a + b + c })
        # ]]]
        options.add_to_whitelist = lowercase_list(normalize_address_list(options.add_to_whitelist))
        options.add_to_blacklist = lowercase_list(normalize_address_list(options.add_to_blacklist))
        options.add_to_unwhitelist = lowercase_list(normalize_address_list(options.add_to_unwhitelist))
        options.add_to_unblacklist = lowercase_list(normalize_address_list(options.add_to_unblacklist))
        options.del_from_whitelist = lowercase_list(normalize_address_list(options.del_from_whitelist))
        options.del_from_blacklist = lowercase_list(normalize_address_list(options.del_from_blacklist))
        options.del_from_unwhitelist = lowercase_list(normalize_address_list(options.del_from_unwhitelist))
        options.del_from_unblacklist = lowercase_list(normalize_address_list(options.del_from_unblacklist))
        # [[[end]]] (checksum: 3a604e2471bc3a3eba01db731048ac26)
        if options.action == 'save':
            options.action = 'text'
        log.debug("Preprocessed: %s", options)

    def parse_args(self, argv):
        """ Parse command line arguments.
            Returns (options, args).
        """
        parser = OptionParser()
        parser.description = "Utility for managing SpamAssassin software"

        parser.add_option('--enable', action='store_true',
                            help=   "Enable spam filtering for all registered mailnames")

        parser.add_option('--disable', action='store_true',
                            help=   "Disable spam filtering for all registered mailnames")


        parser.add_option('-c', '--enable-server-configs', action='store_true',
                            help=   "Enable server configs in filtering process")

        parser.add_option('-C', '--enable-user-configs', action='store_true',
                            help=   "Enable user configs in filtering process")


        parser.add_option('-e', '--enable-mailname', action='store_true',
                            help=   "Add spam filtering for specified mailname")

        parser.add_option('-E', '--disable-mailname', action='store_true',
                            help=   "Remove spam filtering for specified mailname")

        parser.add_option('--enable-checking', action='store_true',
                            help=   "Enable spam filtering for specified mailname")

        parser.add_option('--disable-checking', action='store_true',
                            help=   "Disable spam filtering for specified mailname")


        parser.add_option('-n', '--mailname', action='append',
                            help=   "Perform action on a specified mailname or mailnames list")


        parser.add_option('--max-children', type='int',
                            help=   "Specifies maximum number of child processes. "
                                    "Default is 1 on VZ, 5 otherwise.")

        parser.add_option('-z', '--action', type='choice', 
                            choices=['save', 'text', 'delete', 'move'], default='save',
                            help=   "What to do if a mail is recognized as spam. "
                                    "'save' or 'text' places marked spam mail in mailbox, "
                                    "'delete' removes spam mail without placing it in mailbox "
                                    "and 'move' mode places marked spam mail into 'Spam' "
                                    "mailbox folder. "
                                    "Default value is '%default'.")

        parser.add_option('-r', '--reset-config', action='store_true',
                            help=   "First reset any directives potentially customized by "
                                    "this utility in any SpamAssassin configuration file "
                                    "that is edited. This is especially useful to remove "
                                    "customized black/whitelist settings.")

        parser.add_option('-s', '--sensitivity', type='float',
                            help=   "Set the number of hits required before a mail is "
                                    "considered as spam")

        parser.add_option('-m', '--modify-header', metavar='TAG',
                            help=   "The subject lines of suspected spam will be tagged "
                                    "by this string")

        parser.add_option('-M', '--not-modify-header', 
                            dest='modify_header', action='store_const', const="",
                            help=   "Do not tag subject lines of suspected spam mail")


        parser.add_option('-a', '--add-to-whitelist', action='append', metavar='PATTERN',
                            help=   "Add an email pattern(s) to a whitelist")

        parser.add_option('-A', '--add-to-blacklist', action='append', metavar='PATTERN',
                            help=   "Add an email pattern(s) to a blacklist")

        parser.add_option('-d', '--del-from-whitelist', action='append', metavar='PATTERN',
                            help=   "Delete an email pattern(s) from a whitelist")

        parser.add_option('-D', '--del-from-blacklist', action='append', metavar='PATTERN',
                            help=   "Delete an email pattern(s) from a blacklist")

        parser.add_option('-u', '--add-to-unwhitelist', action='append', metavar='PATTERN',
                            help=   "Add an email pattern(s) to an unwhitelist "
                                    "(remove an email patterns(s) from whitelist)")

        parser.add_option('-U', '--add-to-unblacklist', action='append', metavar='PATTERN',
                            help=   "Add an email pattern(s) to an unblacklist "
                                    "(remove an email pattern(s) from blacklist)")

        parser.add_option('-l', '--del-from-unwhitelist', action='append', metavar='PATTERN',
                            help=   "Delete an email pattern(s) from an unwhitelist "
                                    "(drop removal of an email pattern(s) from whitelist)")

        parser.add_option('-L', '--del-from-unblacklist', action='append', metavar='PATTERN',
                            help=   "Delete an email pattern(s) from an unblacklist "
                                    "(drop removal of an email pattern(s) from blacklist)")


        parser.add_option('--start', action='store_const', const='start', 
                            dest='service_action',
                            help=   "Start the SpamAssassin's daemon")
        
        parser.add_option('--stop', action='store_const', const='stop', 
                            dest='service_action',
                            help=   "Stop the SpamAssassin's daemon")

        parser.add_option('--restart', action='store_const', const='restart', 
                            dest='service_action',
                            help=   "Restart the SpamAssassin's daemon")

        parser.add_option('--condrestart', action='store_const', const='condrestart',
                            dest='service_action',
                            help=   "Restart the SpamAssassin's daemon "
                                    "if it is already running")

        parser.add_option('--status', action='store_const', const='status',
                            dest='service_action',
                            help=   "Check whether the SpamAssassin's daemon is running")

        parser.add_option('--update', action='store_const', const='update',
                            dest='service_action',
                            help=   "Alter the SpamAssassin's daemon configuration "
                                    "without restart")


        options, args = parser.parse_args(args=argv)
        log.debug("Options: %s", options)
        if len(argv) == len(args):
            parser.error("No options were specified. I don't know what to do. "
                         "Call 'spammng --help' for a full list of options.")

        self.preprocess_options(options)
        self.check_options_sanity(options, parser.error)
        return options, args

    def main(self, argv):
        """ Entire spammng, but in a callable form. """
        # original spammng also checked caller and SpamAssassin version, but this seems excessive now
        self.options, args = self.parse_args(argv[1:])
        self.execute()

def main():
    """ spammng main entry point for command-line execution. """
    try:
        SpammngApp().main(sys.argv)
    except SystemExit:
        raise
    except Exception, ex:
        log.error('%s', ex)
        log.debug('This exception happened at:', exc_info=sys.exc_info())
        sys.exit(1)

if __name__ == '__main__':
    main()

# vim: ts=4 sts=4 sw=4 et :
